%% Example demonstrating different options for AMGCL
% Note that these choices mirror the names given in the AMGCL
% documentation: https://amgcl.readthedocs.io/en/latest/runtime.html
%clear all
%
%Program that runs all combinations and writes results to a file
%
mrstModule add msrsb incomp coarsegrid spe10 linearsolvers
%% Create txt file
filename = '/home/shomea/m/mlnorhei/Desktop/MasterThesis/MRST_Sintef/mrst-2018a/mrst-2018a/MyTries/ExperimentalResults/SPE10/layers=1-35_AMG_specificOLD'; %-More ?
filenameError = strcat(filename,'_error.txt');
filename_mldivide = '/home/shomea/m/mlnorhei/Desktop/MasterThesis/MRST_Sintef/mrst-2018a/mrst-2018a/MyTries/ExperimentalResults/mldivide.txt';
filename = strcat(filename,'.txt');
fileID = fopen(filename,'w');
fileIDerror = fopen(filenameError,'w');
fileID_mldivide = fopen(filename_mldivide,'a');
fprintf(fileID,'|%10s|%10s|%26s|%13s|%10s|%12s|%5s|%16s|\r\n','Precond','Solver','Coarsening','Relaxation','Iterations','Time','Conv?','Final residual');
fprintf(fileID,'|----------|----------|--------------------------|-------------|----------|------------|-----|----------------|\r\n');
fprintf(fileIDerror,'|%10s|%10s|%26s|%13s|%s\r\n','Precond','Solver','Coarsening','Relaxation','Warning');


results = zeros(7,4,8); % for 7 Krylov solvers, 4 coarsenings and 8 smoothers
hold on
%% Setup problem
simple = false;
hybrid = false;
olympus = false;

if simple
    model = 'Simple';
    % Create n by n grid
    n = 1000;
    h = 1;
    G = cartGrid([n, n, h]);
    G = computeGeometry(G);
    disp(G)
    rock = makeRock(G, 1*darcy, 1);
    W = [];
    W = addWell(W, G, rock, 1, 'type', 'rate', 'val', 1);
    %W = addWell(W, G, rock, 100, 'type', 'rate', 'val', 1);
    W = addWell(W, G, rock, G.cells.num, 'type', 'bhp', 'val', 0);
    
elseif hybrid
    model = 'Hybrid';
    % The simple case, but with the dimensions and well-setup of SPE10
    layerstart = 32;
    layerstop = 38;
    phi = 1; % porosity makes no difference for incompressible single phase flow
    layers = layerstart:layerstop;
    [G, W, ~] = getSPE10setup(layers);
    rock = makeRock(G,1*darcy,phi);
    
elseif olympus
    model = 'Olympus';
    mrstModule add deckformat
    caseNo = 2;
    casename = ['OLYMPUS_', num2str(caseNo)];
    fn = fullfile(mrstDataDirectory(), 'olympus', 'ECLIPSE', casename, [casename, '.DATA']);
    deck = readEclipseDeck(fn);
    deck = convertDeckUnits(deck);

    % Convert grid, rock and wells
    G = initEclipseGrid(deck);
    G = computeGeometry(G(1));

    rock  = initEclipseRock(deck);
    rock  = compressRock(rock, G.cells.indexMap);

    W = processWells(G, rock, deck.SCHEDULE.control(1));
    % Plot the wells + rock/perm
    figure;
    plotCellData(G, rock.poro);
    plotWell(G, W);

else
    model = 'SPE10';
    % Take part of SPE10 benchmark
    %layers = 1; % 1-35 smoother perm., 36-85 rougher and more varying
    layerstart = 1;
    layerstop = 35;
    layers = layerstart:layerstop;
    [G, W, rock] = getSPE10setup(layers);
end

% Set up linear system by calling the incompTPFA solver without a linear
% solver
fluid = initSingleFluid('rho', 0, 'mu', 1);
state = initResSol(G, 0);
state = incompTPFA(state, G, computeTrans(G, rock), fluid, 'Wells', W, 'matrixOutput', true, 'LinSolve', @(A, b) 0*b);
A = state.A;
b = state.rhs;

% Reduce to only wells
[A, b] = eliminateWellEquations(A, b, G.cells.num);

% Solve the system directly using mldivide, taking the time and writing to
% file:
tic();
%userview = memory;
%reference = userview.MemUsedMATLAB;
x_dir = mldivide(A,b);
%userview = memory;
%mem_ml = userview.MemUsedMATLAB-reference;
t_dir = toc();

conv = 'YES';
if (norm(b-A*x_dir) >= 1e-6) 
    conv = 'NO';
end
if olympus 
    layerstring = '';
    caseNo = num2str(caseNo);
    dim = '';
elseif simple
    layerstring = '';
    caseNo = '';
    dim = strcat(num2str(n), 'x', num2str(n), 'x', num2str(h));
else 
    if layerstart == layerstop
        layerstring = num2str(layerstart);
    else
        layerstring = strcat(num2str(layerstart),':',num2str(layerstop));
    end
    caseNo = '';
    dim = '';
end
fprintf(fileID_mldivide,'|%10s|%10s|%10s|%15s|%13.2f|%5s|\r\n',model, caseNo, layerstring, dim, t_dir, conv);


fclose(fileID_mldivide);



% Make a choice between all alternative preconditioners, solvers,
% coarsenings and smoothers
makeChoice = @(text, choices)  listdlg('PromptString',  text,...
                                       'SelectionMode', 'single',...
                                       'ListString',    choices);
nt = 2; % the maximum number of computational threads





%spparms('spumoni',2) 




%% Select combinations:

numcom = 5; % number of combinations

% Combo 1;
c1 = [1,1,5]; % BiCGstab + R-S + ILU(l)
% Combo 2;
c2 = [2,1,5]; % CG + R-S + ILU(l)
% Combo 3;
c3 = [3,1,6]; % BiCGstabl + R-S + ILUT
% Combo 4;
c4 = [3,1,5]; % BiCGstabl + R-S + ILU(l)
% Combo 5;
c5 = [7,1,5]; % IDR(s)  + R-S + ILU(l)
% Combo 6;
%c6 = [4,2,4]; % GMRES + sm. aggr. + ILU(0)
% Combo 7;
%c7 = [5,2,4]; % LGMRES + sm. aggr. + ILU(0)
% Combo 8;
%c8 = [6,2,4]; % FGMRES + sm. aggr. + ILU(0)
% Combo 9;
%c9 = [7,2,4]; % IDR(s) + sm. aggr. + ILU(0)

% All in one matrix:
C = [c1;c2;c3;c4;c5]; %;c6;c7;c8;c9];


%% Select preconditioner
% There are two options here. If relaxation (AMG, you mean?) is chosen, AMG will be used as
% a preconditioner. If relaxation is chosen, no multigrid will be used and
% the solver will be a simpler Krylov solver with relaxation as a
% preconditioner. The coarsening option chosen later in this example will
% not be used in this case.

% Preconditioners
preconditioners = {'amg', ...
                   'relaxation'};
%index = makeChoice('Select solver', preconditioners);


%% Solvers
% Choices for Krylov acceleration
solvers = {'bicgstab', ... % Biconjugate gradient stabilized method
           'cg', ... % Conjugate gradient
           'bicgstabl', ... % Bicgstab variant
           'gmres', ... % Generalized minimal residual method
           'lgmres', ... % GMRES variant
           'fgmres', ... % Flexible-GMRES 
           'idrs'}; % Induced Dimension Reduction method IDR(s)
%index = makeChoice('Select Krylov-solver', solvers);

%% Select coarsening and interpolation scheme for AMG
coarsening = {'ruge_stuben', ... % Ruge-stuben coarsening ("standard" coarsening)
              'smoothed_aggregation', ... % Aggregation with smoothed iterates
              'aggregation', ... % Aggregation - very fast coarsening due to very simple constant interpolation
              'smoothed_aggr_emin'}; % Aggregation with energy-minimizing smoothing
%index = makeChoice('Select coarsening', coarsening);

%% Select relaxation ("smoother")
relaxations = {'spai0', ... % Sparse approximate inverse, degree 0
               'spai1', ... % Sparse approximate inverse
               'gauss_seidel', ... % Gauss-seidel
               'ilu0', ... % Incomplete LU-factorization without fill-in
               'iluk', ... % Level-based ILU
               'ilut', ... % Thresholded ILU
               'damped_jacobi', ... % Damped Jacobi
               'chebyshev'}; % Chebyshev smoothing
%index = makeChoice('Select relaxation', relaxations);
%relax = relaxations{index};

%%%%%%%%

%fprintf('============================================================================ \n %s, %s, %s, %s \n',precond,solver,coarsen,relax)
%% Solve the system (but avoid bad combinations)
%      
%                  if ( strcmp(precond,'amg') && strcmp(coarsen,'smoothed_aggr_emin') && strcmp(relax,'ilu0') )
%                      fprintf('We avoid amg+smoth.aggr.emin.+ilu0 \n \n \n')
%                  elseif ( strcmp(precond,'amg') && strcmp(solver,'bicgstabl') && strcmp(coarsen,'smoothed_aggr_emin') )
%                      fprintf('We avoid amg+bicgstabl+smooth.aggr.emin. \n \n \n')
%                  elseif ( strcmp(precond,'amg') && strcmp(solver,'idrs') && strcmp(coarsen,'smoothed_aggr_emin') )
%                      fprintf('We avoid amg+idrs+smooth.aggr.emin. \n \n \n')
%                  elseif ( strcmp(precond,'amg') && strcmp(solver,'bicgstab') && strcmp(coarsen,'ruge_stuben') && strcmp(relax,'spai1') )
%                      fprintf('We avoid amg+bicgstab+ruge_stuben+spai1 \n \n \n')
%                  if ( strcmp(precond,'relaxation') && strcmp(solver,'bicgstab') && (strcmp(relax,'spai1') || strcmp(relax,'ilu0')) )
%                      fprintf('We avoid relaxation+bicgstab+ilu0 \n \n \n')
%                  else      
%                     tic();
%                     maxNumCompThreads(nt)
%                     maxIter = 5000;
%                     [x, err, iter] = callAMGCL(A, b, ... 
%                                         'coarsening', coarsen,...
%                                         'relaxation', relax, ...
%                                         'solver', solver, ...
%                                         'maxIterations', maxIter, ...
%                                         'preconditioner', precond, ...
%                                         'verbose', true);
%                     t_amg = toc();
%                  end

precond = preconditioners{1}; % AMG=1, relaxation=2

for combos = 1:numcom
    currectCombo = C(combos,:);
    solver = solvers{currectCombo(1)};
    coarsen = coarsening{currectCombo(2)};
    relax = relaxations{currectCombo(3)};

    try
        tic();
        maxNumCompThreads(nt)
        maxIter = 5000;
        [x, err, iter] = callAMGCL(A, b, ... 
                           'coarsening', coarsen,...
                           'relaxation', relax, ...
                           'solver', solver, ...
                           'maxIterations', maxIter, ...
                           'preconditioner', precond, ...
                           'verbose', true);
        %userview = memory;
        %mem = userview.MemUsedMATLAB-reference;
        t_amg = toc();
    catch ME
        message = ME.message;
        fprintf(fileIDerror,'|%10s|%10s|%26s|%13s|\r\n%s\r\n',precond,solver,coarsen,relax,message);
        %ME
        %if ~simple
        %    [x,err,iter] = zeros((60*220));0;0;
        %else
        %    [x,err,iter] = zeros((n*n));0;0;
        %end
    end
    %% Results
    conv = 'YES';
    if (iter==5000) || (norm(b-A*x_dir) >= 1e-6) 
        conv = 'NO';
    end
    fprintf(fileID,'|%10s|%10s|%26s|%13s|%10i|%12.2f|%5s|%16d|\r\n',precond,solver,coarsen,relax,iter,t_amg,conv,err);

    fprintf(fileID,'|----------|----------|--------------------------|-------------|----------|------------|-----|----------------|\r\n');

end
         
         
         





%% Additional options
% There are many additional options which can be set through the AMGCL
% struct. Generally, these can be passed as keyword arguments to callAMGCL
% and other routines which rely on AMGCL.
% 
% Especially interesting options are
%  - ncycle (number of multigrid cycles)
%  - npre (number of presmootings)
%  - npost (number of postsmoothings)
% As well as options relating to the coarsening hierarchy:
%  - rs_eps_strong (strong threshold, ruge_stuben)
%  - aggr_eps_strong (strong threshold, aggregation)
%str = getAMGCLMexStruct();
%disp(str)